from __future__ import absolute_import

from os.path import join, abspath, dirname

PROJECT_ROOT = abspath(dirname(__file__))

APP_SETTINGS = dict(
    template_path=join(PROJECT_ROOT, "templates"),
    static_path=join(PROJECT_ROOT, "static"),
    debug=True,
)

PAGE_UPDATE_INTERVAL = 2000
CELERY_INSPECT_INTERVAL = 1000



from __future__ import absolute_import

import logging
import threading

import celery

from celery.events import EventReceiver
from celery.messaging import establish_connection
from celery.events.state import state
from celery.events.snapshot import Polaroid

celery = celery.current_app


class DumpCam(Polaroid):
    def __init__(self, state, freq=1.0):
        Polaroid.__init__(self, state, freq)

    def on_shutter(self, state):
        if not state.event_count:
            # No new events since last snapshot.
            return


class EventCollector(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        logging.info("Enabling events")
        celery.control.enable_events()

        while True:
            try:
                with establish_connection() as connection:
                    recv = EventReceiver(
                            connection, handlers={"*": state.event})
                    with DumpCam(state, freq=1.0):
                        recv.capture(limit=None, timeout=None)
            except (KeyboardInterrupt, SystemExit):
                import thread
                thread.interrupt_main()
            except Exception as e:
                logging.error("An error occurred while capturing events"
                              ": %s" % e)


tasks = state.tasks



from __future__ import absolute_import

import logging

import celery

from tornado import web

from ..views import BaseHandler
from ..models import WorkersModel

is_worker = WorkersModel.is_worker
celery = celery.current_app


class ShutdownWorker(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        logging.info("Shutting down '%s' worker" % workername)
        celery.control.broadcast('shutdown', destination=[workername])
        self.write(dict(message="Shutting down!"))


class RestartWorkerPool(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        logging.info("Restarting '%s' worker's pool" % workername)
        response = celery.control.broadcast('pool_restart',
                                             arguments={'reload': False},
                                             destination=[workername],
                                             reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(
                message="Restarting '%s' worker's pool" % workername))
        else:
            logging.error(response)
            self.set_status(403)
            self.write("Failed to restart the '%s' pool" % workername)


class TaskRateLimit(BaseHandler):
    def post(self, workername=None):
        if workername is not None and not is_worker(workername):
            raise web.HTTPError(404)

        taskname = self.get_argument('taskname', None)
        ratelimit = int(self.get_argument('ratelimit'))

        logging.info("Setting '%s' rate limit for '%s' task" %
                     (ratelimit, taskname))
        response = celery.control.rate_limit(taskname,
                                              ratelimit,
                                              reply=True,
                                              destination=[workername])
        if 'ok' in response[0][workername]:
            self.write(dict(message=response[0][workername]['ok']))
        else:
            logging.error(response)
            self.set_status(403)
            self.write("Failed to set rate limit: '%s'" %
                       response[0][workername]['error'])


class TaskTimout(BaseHandler):
    def post(self, workername=None):
        if workername is not None and not is_worker(workername):
            raise web.HTTPError(404)

        taskname = self.get_argument('taskname', None)
        hard = self.get_argument('hard-timeout', None)
        soft = self.get_argument('soft-timeout', None)
        hard = hard and float(hard)
        soft = soft and float(soft)

        logging.info("Setting timeouts for '%s' task" % taskname)
        response = celery.control.time_limit(taskname, hard, soft,
                                             reply=True,
                                             destination=[workername])
        if 'ok' in response[0][workername]:
            self.write(dict(message=response[0][workername]['ok']))
        else:
            logging.error(response)
            self.set_status(403)
            self.write("Failed to set timeouts: '%s'" %
                       response[0][workername]['error'])


class WorkerPoolGrow(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        n = int(self.get_argument('n', 1))

        logging.info("Growing '%s' worker's pool by '%s'" % (workername, n))
        response = celery.control.broadcast('pool_grow',
                                             arguments={'n': n},
                                             destination=[workername],
                                             reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(message="Growing '%s' worker's pool" % workername))
        else:
            logging.error(response)
            self.set_status(403)
            self.write("Failed to grow '%s' worker's pool" % workername)


class WorkerPoolShrink(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        n = int(self.get_argument('n', 1))

        logging.info("Shrinking '%s' worker's pool by '%s'" % (workername, n))
        response = celery.control.broadcast('pool_shrink',
                                             arguments={'n': n},
                                             destination=[workername],
                                             reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(
                message="Shrinking '%s' worker's pool" % workername))
        else:
            logging.error(response)
            self.set_status(403)
            self.write("Failed to restart '%s' worker's pool" % workername)


class WorkerPoolAutoscale(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        min = int(self.get_argument('min'))
        max = int(self.get_argument('max'))

        logging.info("Autoscaling '%s' worker by '%s'" %
                     (workername, (min, max)))
        response = celery.control.broadcast('autoscale',
                        arguments={'min': min, 'max': max},
                        destination=[workername],
                        reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(message="Autoscaling '%s' worker" % workername))
        else:
            logging.error(response)
            error = response[0][workername]['error']
            self.set_status(403)
            self.write("Failed to autoscale '%s' worker: %s" %
                       (workername, error))


class WorkerQueueAddConsumer(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        queue = self.get_argument('queue')

        logging.info("Adding consumer '%s' to worker '%s'" %
                     (queue, workername))
        response = celery.control.broadcast('add_consumer',
                        arguments={'queue': queue},
                        destination=[workername],
                        reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(message=response[0][workername]['ok']))
        else:
            logging.error(response)
            error = response[0][workername]['error']
            self.set_status(403)
            self.write("Failed to add '%s' consumer to '%s' worker: %s" %
                       (queue, workername, error))


class WorkerQueueCancelConsumer(BaseHandler):
    def post(self, workername):
        if not is_worker(workername):
            raise web.HTTPError(404)

        queue = self.get_argument('queue')

        logging.info("Canceling consumer '%s' from worker '%s'" %
                     (queue, workername))
        response = celery.control.broadcast('cancel_consumer',
                        arguments={'queue': queue},
                        destination=[workername],
                        reply=True)
        if 'ok' in response[0][workername]:
            self.write(dict(message=response[0][workername]['ok']))
        else:
            logging.error(response)
            error = response[0][workername]['error']
            self.set_status(403)
            self.write("Failed to cancel '%s' consumer from '%s' worker: %s" %
                       (queue, workername, error))



from __future__ import absolute_import

import logging

from tornado import websocket
from tornado.ioloop import PeriodicCallback

from ..models import WorkersModel
from ..settings import PAGE_UPDATE_INTERVAL


class UpdateWorkers(websocket.WebSocketHandler):
    listeners = []
    periodic_callback = None
    workers = None

    def open(self):
        listeners = UpdateWorkers.listeners
        periodic_callback = UpdateWorkers.periodic_callback

        if not listeners:
            periodic_callback = periodic_callback or PeriodicCallback(
                    UpdateWorkers.on_update_time, PAGE_UPDATE_INTERVAL)
            periodic_callback.start()
        listeners.append(self)

    def on_message(self, message):
        logging.error(message)

    def on_close(self):
        listeners = UpdateWorkers.listeners
        periodic_callback = UpdateWorkers.periodic_callback

        listeners.remove(self)
        if not listeners and periodic_callback:
            periodic_callback.stop()

    @classmethod
    def on_update_time(cls):
        workers = WorkersModel.get_latest()

        if workers != cls.workers:
            for l in cls.listeners:
                l.write_message(workers.workers)
            cls.workers = workers



from __future__ import absolute_import

from tornado import web

from ..views import BaseHandler
from ..models import TaskModel, WorkersModel


class TaskView(BaseHandler):
    def get(self, task_id):
        task = TaskModel.get_task_by_id(task_id)
        if task is None:
            raise web.HTTPError(404)

        self.render("task.html", task=task)


class TasksView(BaseHandler):
    def get(self):
        limit = self.get_argument('limit', None)
        worker = self.get_argument('worker', None)
        type = self.get_argument('type', None)

        limit = limit and int(limit)
        worker = worker if worker != 'All' else None
        type = type if type != 'All' else None

        tasks = TaskModel.iter_tasks(limit=limit, type=type, worker=worker)
        workers = WorkersModel.get_workers()
        seen_task_types = TaskModel.seen_task_types()

        self.render("tasks.html", tasks=tasks,
                                  task_types=seen_task_types,
                                  workers=workers,
                                  limit=limit,
                                  worker=worker,
                                  type=type)



from __future__ import absolute_import

import inspect

import tornado

from ..utils import template


class BaseHandler(tornado.web.RequestHandler):
    def render(self, *args, **kwargs):
        functions = inspect.getmembers(template, inspect.isfunction)
        assert not set(map(lambda x: x[0], functions)) & set(kwargs.iterkeys())
        kwargs.update(functions)
        super(BaseHandler, self).render(*args, **kwargs)



from __future__ import absolute_import

import re

from datetime import datetime

KEYWORDS_UP = ('ssl', 'uri', 'url', 'uuid', 'eta')
KEYWORDS_DOWN = ('args', 'kwargs')
UUID_REGEX = re.compile(r'^[\w]{8}(-[\w]{4}){3}-[\w]{12}$')


def humanize(obj, type=None):
    if obj is None:
        obj = ''
    elif type == 'time':
        obj = datetime.fromtimestamp(float(obj)).strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(obj, basestring) and not re.match(UUID_REGEX, obj):
        obj = obj.replace('-', ' ').replace('_', ' ')
        obj = re.sub('|'.join(KEYWORDS_UP),
                     lambda m: m.group(0).upper(), obj)
        if obj and obj not in KEYWORDS_DOWN:
            obj = obj[0].upper() + obj[1:]
    elif isinstance(obj, list):
        if all(map(lambda x: isinstance(x, (int, float, basestring)), obj)):
            obj = ', '.join(map(str, obj))
    return obj



from __future__ import absolute_import
from __future__ import with_statement

import time
import copy
import logging
import threading

import celery

from .settings import CELERY_INSPECT_INTERVAL


class State(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.daemon = True

        self._update_lock = threading.Lock()
        self._stats = {}
        self._registered_tasks = {}
        self._scheduled_tasks = {}
        self._active_tasks = {}
        self._reserved_tasks = {}
        self._revoked_tasks = {}
        self._ping = {}
        self._active_queues = {}

    def run(self):
        i = celery.current_app.control.inspect()
        while True:
            try:
                stats = i.stats() or {}
                registered = i.registered() or {}
                scheduled = i.scheduled() or {}
                active = i.active() or {}
                reserved = i.reserved() or {}
                revoked = i.revoked() or {}
                ping = i.ping() or {}
                active_queues = i.active_queues() or {}

                with self._update_lock:
                    self._stats = stats
                    self._registered_tasks = registered
                    self._scheduled_tasks = scheduled
                    self._active_tasks = active
                    self._reserved_tasks = reserved
                    self._revoked_tasks = revoked
                    self._ping = ping
                    self._active_queues = active_queues

                time.sleep(CELERY_INSPECT_INTERVAL / 1000)
            except (KeyboardInterrupt, SystemExit):
                import thread
                thread.interrupt_main()
            except Exception as e:
                logging.error("An error occurred while inspecting workers"
                              ": %s" % e)


    @property
    def stats(self):
        with self._update_lock:
            return copy.deepcopy(self._stats)

    @property
    def registered_tasks(self):
        with self._update_lock:
            return copy.deepcopy(self._registered_tasks)

    @property
    def scheduled_tasks(self):
        with self._update_lock:
            return copy.deepcopy(self._scheduled_tasks)

    @property
    def active_tasks(self):
        with self._update_lock:
            return copy.deepcopy(self._active_tasks)

    @property
    def reserved_tasks(self):
        with self._update_lock:
            return copy.deepcopy(self._reserved_tasks)

    @property
    def revoked_tasks(self):
        with self._update_lock:
            return copy.deepcopy(self._revoked_tasks)

    @property
    def ping(self):
        with self._update_lock:
            return copy.deepcopy(self._ping)

    @property
    def active_queues(self):
        with self._update_lock:
            return copy.deepcopy(self._active_queues)


state = State()



from __future__ import absolute_import

from tornado import ioloop
from tornado.web import Application
from tornado.options import define, options, parse_command_line

from .state import state
from .events import EventCollector
from .urls import handlers
from .settings import APP_SETTINGS

define("port", default=8008, help="run on the given port", type=int)


def main(argv=None):
    parse_command_line(argv)

    application = Application(handlers, **APP_SETTINGS)

    print('> visit me at http://localhost:%s' % options.port)

    application.listen(options.port)
    try:
        state.start()
        EventCollector().start()
        ioloop.IOLoop.instance().start()
    except (KeyboardInterrupt, SystemExit):
        pass


if __name__ == "__main__":
    main()



from __future__ import absolute_import

from tornado import web

from ..views import BaseHandler
from ..models import WorkersModel, WorkerModel


class WorkersView(BaseHandler):
    def get(self):
        self.render("workers.html",
                    workers=WorkersModel.get_latest().workers)


class WorkerView(BaseHandler):
    def get(self, workername):
        worker = WorkerModel.get_worker(workername)
        if worker is None:
            raise web.HTTPError(404)

        self.render("worker.html", worker=worker)



from __future__ import absolute_import

from tornado.web import StaticFileHandler

from .views.worker import (
        WorkersView,
        WorkerView,
        )

from .views.tasks import (
        TaskView,
        TasksView,
        )

from .views.control import (
        ShutdownWorker,
        RestartWorkerPool,
        WorkerPoolGrow,
        WorkerPoolShrink,
        WorkerPoolAutoscale,
        WorkerQueueAddConsumer,
        WorkerQueueCancelConsumer,
        TaskRateLimit,
        TaskTimout,
        )

from .views.update import (
        UpdateWorkers,
        )

from .settings import APP_SETTINGS


handlers = [
    # App
    (r"/", WorkersView),
    (r"/workers", WorkersView),
    (r"/worker/(.+)", WorkerView),
    (r"/task/(.+)", TaskView),
    (r"/tasks", TasksView),
    # Commands
    (r"/shut-down-worker/(.+)", ShutdownWorker),
    (r"/restart-pool-worker/(.+)", RestartWorkerPool),
    (r"/worker-pool-grow/(.+)", WorkerPoolGrow),
    (r"/worker-pool-shrink/(.+)", WorkerPoolShrink),
    (r"/worker-pool-autoscale/(.+)", WorkerPoolAutoscale),
    (r"/worker-queue-add-consumer/(.+)", WorkerQueueAddConsumer),
    (r"/worker-queue-cancel-consumer/(.+)", WorkerQueueCancelConsumer),
    (r"/task-timeout/(.+)", TaskTimout),
    (r"/task-rate-limit/(.+)", TaskRateLimit),
    # WebSocket Updates
    (r"/update-workers", UpdateWorkers),
    # Static
    (r"/(.*)", StaticFileHandler, {"path": APP_SETTINGS['static_path']})
]



#!/usr/bin/env python
import os
import re

from setuptools import setup, find_packages


version = re.compile(r'VERSION\s*=\s*\((.*?)\)')


def get_package_version():
    "returns package version without importing it"
    base = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base, "flower/__init__.py")) as initf:
        for line in initf:
            m = version.match(line.strip())
            if not m:
                continue
            return ".".join(m.groups()[0].split(", "))


setup(
    name='flower',
    version=get_package_version(),
    description='Celery Flower',
    long_description=open('README.rst').read(),
    author='Mher Movsisyan',
    url='https://github.com/mher/flower',
    packages=find_packages(),
    install_requires=['celery', 'tornado'],
    package_data={'flower': ['templates/*', 'static/**/*']},
    entry_points={
        'console_scripts': [
            'flower = flower.__main__:main',
        ],
        'celery.commands': [
            'flower = flower.command:Admin',
        ],
    },
)



from __future__ import absolute_import
from __future__ import with_statement

from collections import OrderedDict

from .state import state
from .events import tasks
from .events import state as event_state


class BaseModel(object):
    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)


class WorkersModel(BaseModel):
    def __init__(self, state):
        super(WorkersModel, self).__init__()
        self.workers = OrderedDict()
        self.initialize(state)

    def initialize(self, state):
        for workername, stat in sorted(state.stats.iteritems()):
            self.workers[workername] = dict(
                    status=(workername in state.ping),
                    concurrency=stat['pool']['max-concurrency'],
                    completed_tasks=sum(stat['total'].itervalues()),
                    running_tasks=len(state.active_tasks[workername]),
                    queues=map(lambda x: x['name'],
                               state.active_queues[workername]),
                    )

    @classmethod
    def get_latest(cls):
        return WorkersModel(state)

    @classmethod
    def get_workers(cls):
        return state.stats.keys()

    @classmethod
    def is_worker(cls, workername):
        return WorkerModel(workername, state) is not None

    def __eq__(self, other):
        return other is not None and self.workers == other.workers


class WorkerModel(BaseModel):
    def __init__(self, workername, state):
        super(BaseModel, self).__init__()
        self.initialize(workername, state)

    def initialize(self, name, state):
        self.name = name
        self.stats = state.stats[name]
        self.active_tasks = state.active_tasks[name]
        self.scheduled_tasks = state.scheduled_tasks[name]
        self.active_queues = state.active_queues[name]
        self.revoked_tasks = state.revoked_tasks[name]
        self.registered_tasks = filter(lambda x: not x.startswith('celery.'),
                                       state.registered_tasks[name])
        self.reserved_tasks = state.reserved_tasks[name]

    @classmethod
    def get_worker(self, name):
        if name not in state.stats:
            return None
        return WorkerModel(name, state)

    def __eq__(self, other):
        return self.name == other.name and self.stats == other.stats and\
               self.active_tasks == other.active_tasks and\
               self.active_queues == other.active_queues and\
               self.revoked_tasks == other.revoked_tasks and\
               self.registered_tasks == other.registered_tasks and\
               self.scheduled_tasks == other.scheduled_tasks and\
               self.reserved_tasks == other.reserved_tasks


class TaskModel(BaseModel):
    def __init__(self, task_id):
        super(BaseModel, self).__init__()

        task = tasks[task_id]

        self._fields = task._defaults.keys()
        for name, value in task.info(fields=self._fields).iteritems():
            setattr(self, name, value)

    @classmethod
    def get_task_by_id(cls, task_id):
        try:
            return TaskModel(task_id)
        except KeyError:
            return None

    @classmethod
    def iter_tasks(cls, limit=None, type=None, worker=None):
        i = 0
        for uuid, task in event_state._sort_tasks_by_time(
                event_state.itertasks()):
            if type and task.name != type:
                continue
            if worker and task.worker.hostname != worker:
                continue
            yield uuid, task
            i += 1
            if i == limit:
                break

    @classmethod
    def seen_task_types(cls):
        return event_state.task_types()

    def __dir__(self):
        return self._fields



from celery.bin.base import Command

from .__main__ import main


class Admin(Command):

    def run_from_argv(self, prog_name, argv=None):
        argv = self.setup_app_from_commandline(argv)
        return main(argv)

